/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 1996, 1998, 1999, 2004
 *	Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SUDOERS_INS_CSOPS_H
#define SUDOERS_INS_CSOPS_H

    /*
     * CSOps insults (may be site dependent).
     */

    N_("Maybe if you used more than just two fingers..."),
    N_("BOB says:  You seem to have forgotten your passwd, enter another!"),
    N_("stty: unknown mode: doofus"),
    N_("I can't hear you -- I'm using the scrambler."),
    N_("The more you drive -- the dumber you get."),
    N_("Listen, broccoli brains, I don't have time to listen to this trash."),
    N_("I've seen penguins that can type better than that."),
    N_("Have you considered trying to match wits with a rutabaga?"),
    N_("You speak an infinite deal of nothing"),

#endif /* SUDOERS_INS_CSOPS_H */
