/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 2018 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SUDOERS_INS_PYTHON_H
#define SUDOERS_INS_PYTHON_H

    /*
     * Insults from "Monty Python's Flying Circus" and family.
     */

    N_("That is no basis for supreme executive power!"),
    N_("You empty-headed animal food trough wiper!"),
    N_("I fart in your general direction!"),
    N_("Your mother was a hamster and your father smelt of elderberries!"),
    N_("You must cut down the mightiest tree in the forest... with... a herring!"),
    N_("I wave my private parts at your aunties!"),
    N_("He's not the Messiah, he's a very naughty boy!"),
    N_("I wish to make a complaint."),
    N_("When you're walking home tonight, and some homicidal maniac comes after you with a bunch of loganberries, don't come crying to me!"),
    N_("This man, he doesn't know when he's beaten! He doesn't know when he's winning, either. He has no... sort of... sensory apparatus..."),
    N_("There's nothing wrong with you that an expensive operation can't prolong."),
    N_("I'm very sorry, but I'm not allowed to argue unless you've paid."),

#endif /* SUDOERS_INS_PYTHON_H */
