/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 1996, 1998, 1999 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SUDOERS_INS_2001_H
#define SUDOERS_INS_2001_H

    /*
     * HAL insults (paraphrased) from 2001.
     */

    N_("Just what do you think you're doing Dave?"),
    N_("It can only be attributed to human error."),
    N_("That's something I cannot allow to happen."),
    N_("My mind is going. I can feel it."),
    N_("Sorry about this, I know it's a bit silly."),
    N_("Take a stress pill and think things over."),
    N_("This mission is too important for me to allow you to jeopardize it."),
    N_("I feel much better now."),

#endif /* SUDOERS_INS_2001_H */
