/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 1996, 1998, 1999 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SUDOERS_INS_GOONS_H
#define SUDOERS_INS_GOONS_H

    /*
     * Insults from the "Goon Show."
     */

    N_("You silly, twisted boy you."),
    N_("He has fallen in the water!"),
    N_("We'll all be murdered in our beds!"),
    N_("You can't come in. Our tiger has got flu"),
    N_("I don't wish to know that."),
    N_("What, what, what, what, what, what, what, what, what, what?"),
    N_("You can't get the wood, you know."),
    N_("You'll starve!"),
    N_("... and it used to be so popular..."),
    N_("Pauses for audience applause, not a sausage"),
    N_("Hold it up to the light --- not a brain in sight!"),
    N_("Have a gorilla..."),
    N_("There must be cure for it!"),
    N_("There's a lot of it about, you know."),
    N_("You do that again and see what happens..."),
    N_("Ying Tong Iddle I Po"),
    N_("Harm can come to a young lad like that!"),
    N_("And with that remarks folks, the case of the Crown vs yourself was proven."),
    N_("Speak English you fool --- there are no subtitles in this scene."),
    N_("You gotta go owwwww!"),
    N_("I have been called worse."),
    N_("It's only your word against mine."),
    N_("I think ... err ... I think ... I think I'll go home"),

#endif /* SUDOERS_INS_GOONS_H */
