/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 2020 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * This is an open source non-commercial project. Dear PVS-Studio, please check it.
 * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
 */

#include <config.h>

#include <errno.h>
#include <unistd.h>

#include <sudo_compat.h>

#if !defined(HAVE_PWRITE) && !defined(HAVE_PWRITE64)
ssize_t
sudo_pwrite(int fd, const void *buf, size_t nbytes, off_t offset)
{
    ssize_t nwritten;
    off_t old_offset;

    old_offset = lseek(fd, (off_t)0, SEEK_CUR);
    if (old_offset == -1 || lseek(fd, offset, SEEK_SET) == -1)
	return -1;

    nwritten = write(fd, buf, nbytes);
    if (lseek(fd, old_offset, SEEK_SET) == -1)
	return -1;

    return nwritten;
}
#endif /* HAVE_PWRITE && !HAVE_PWRITE64 */
